"""
   Basic Encoding Rules implementation for "universal" ASN.1 data types.

   Copyright 1999-2002 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.

   The BER processing part of this code has been partially derived
   from Simon Leinen's <simon@switch.ch> BER Perl module.
"""
# Module public names
__all__ = [ 'Boolean', 'Integer', 'BitString', 'OctetString', 'Null', \
            'ObjectIdentifier', 'Real', 'Enumerated', 'Sequence', \
            'SequenceOf', 'Set', 'SetOf', 'Choice' ]

import string
from pysnmp.asn1.ber import base, error

class Boolean(base.SimpleAsn1Object):
    """Implements BER processing for an ASN.1 boolean object
    """
    def _encode(self, value):
        """Encode boolean value
        """
        if value:
            return '\377'
        else:
            return '\000'

    def _decode(self, input):
        """Decode input into boolean value
        """
        if len(input) != 1:
            raise error.BadEncodingError('Input of wrong size (%d) for %s' %\
                                         (len(input), self.__class__.__name__))
        if ord(input[0]):
            return 1
        else:
            return 0

class Integer(base.SimpleAsn1Object):
    """Implements BER processing for an ASN.1 integer object
    """
    def _encode(self, integer):
        """
           _encode() -> octet stream
           
           Encode tagged integer into octet stream.
        """
        result = ''
        
        # The 0 and -1 values need to be handled separately since
        # they are the terminating cases of the positive and negative
        # cases repectively.
        if integer == 0:
            result = '\000'
            
        elif integer == -1:
            result = '\377'
            
        elif integer < 0:
            while integer <> -1:
                (integer, result) = integer>>8, chr(integer & 0xff) + result
                
            if ord(result[0]) & 0x80 == 0:
                result = chr(0xff) + result
        else:
            while integer > 0:
                (integer, result) = integer>>8, chr(integer & 0xff) + result
                
            if (ord(result[0]) & 0x80 <> 0):
                result = chr(0x00) + result

        return result

    def _decode(self, input):
        """
           _decode(input) -> IntValue
           
           Decode octet stream into signed ASN.1 integer (of any length).
        """
        bytes = map(ord, input)

        if bytes[0] & 0x80:
            bytes.insert(0, -1L)

        result = reduce(lambda x,y: x<<8 | y, bytes, 0L)

        try:
            return int(result)

        except OverflowError:
            return result

class BitString(base.SimpleAsn1Object):
    """BER for ASN.1 BITSTRING type
    """
    pass

class OctetString(base.SimpleAsn1Object):
    """Implements BER processing for an ASN.1 OCTETSTRING object
    """
    def _encode(self, value):
        """Encode octet string value
        """
        return value

    def _decode(self, input):
        """Decode input into octet string value
        """
        return input

class Null(base.SimpleAsn1Object):
    """BER for ASN.1 NULL object
    """
    def _encode(self, value):
        """Encode ASN.1 NULL value
        """
        return ''

    def _decode(self, input):
        """Decode input into ASN.1 NULL value
        """
        if len(input) != 0:
            raise error.BadEncodingError('Input of wrong size (%d) for %s' %\
                                         (len(input), self.__class__.__name__))
        return ''

class ObjectIdentifier(base.SimpleAsn1Object):
    """BER for ASN.1 OBJECTIDENTIFIER object
    """
    def _encode(self, oid):
        """
           _encode() -> octet stream
           
           Encode ASN.1 Object ID into octet stream.
        """
        # Make sure the Object ID is long enough
        if len(oid) < 2:
            raise error.BadArgumentError('Short Object ID for %s' % \
                                         self.__class__.__name__)

        # Build the first twos
        index = 0
        result = oid[index] * 40
        result = result + oid[index+1]
        try:
            result = [ '%c' % int(result) ]

        except OverflowError:
            raise error.BadArgumentError('Too large initial sub-IDs %s for %s' %\
                                         (str(oid[index:]), \
                                    self.__class__.__name__))

        # Setup index
        index = index + 2

        # Cycle through subids
        for subid in oid[index:]:
            if subid > -1 and subid < 128:
                # Optimize for the common case
                result.append('%c' % (subid & 0x7f))

            elif subid < 0 or subid > 0xFFFFFFFFL:
                raise error.BadArgumentError('Too large Sub-Object ID %s for %s' %\
                                             (str(subid), self.__class__.__name__))

            else:
                # Pack large Sub-Object IDs
                res = [ '%c' % (subid & 0x7f) ]
                subid = subid >> 7
                while subid > 0:
                    res.insert(0, '%c' % (0x80 | (subid & 0x7f)))
                    subid = subid >> 7

                # Convert packed Sub-Object ID to string and add packed
                # it to resulted Object ID
                result.append(string.join(res, ''))

        # Convert BER encoded Object ID to string and return
        return string.join(result, '')
        
    def _decode(self, input):
        """
           _decode(input)
           
           Decode octet stream into ASN.1 Object ID
        """
        oid = []
        index = 0

        if len(input) == 0:
            raise error.BadArgumentError('Short Object ID at %s'
                                         % self.__class__.__name__)
        # Get the first subid
        subid = ord(input[index])
        oid.append(int(subid / 40))
        oid.append(int(subid % 40))

        index = index + 1

        # Loop through the rest
        while index < len(input):
            # Get a subid
            subid = ord(input[index])

            if subid < 128:
                oid.append(subid)
                index = index + 1
            else:
                # Construct subid from a number of octets
                next = subid
                subid = 0
                while next >= 128:
                    # Collect subid
                    subid = (subid << 7) + (next & 0x7F)

                    # Take next octet
                    index = index + 1
                    next = ord(input[index])

                    # Just for sure
                    if index > len(input):
                        raise error.BadArgumentError('Malformed sub-Object ID at %s'
                                                     % self.__class__.__name__)

                # Append a subid to oid list
                subid = (subid << 7) + next
                oid.append(subid)
                index = index + 1

        return oid

class Real(base.SimpleAsn1Object):
    """BER for ASN.1 REAL object
    """
    pass

class Enumerated(Integer):
    """Same as Integer
    """
    pass

# BER for structured ASN.1 objects

class Sequence(base.OrderedFixedTypeAsn1Object):
    """Implements BER processing for an ASN.1 SEQUENCE object
    """
    pass

class SequenceOf(base.OrderedVariableTypeAsn1Object):
    """Implements BER processing for an ASN.1 SEQUENCE OF object
    """
    pass

class Set(base.UnorderedFixedTypeAsn1Object):
    """Implements BER processing for an ASN.1 SET object
    """
    pass

class SetOf(base.UnorderedVariableTypeAsn1Object):
    """Implements BER processing for an ASN.1 SET OF object
    """
    pass

class Choice(base.SingleFixedTypeAsn1Object):
    """Implements BER processing for an ASN.1 CHOICE object
    """
    pass

#
# A table of supported ASN.1 types (XXX: is it needed?)
#

supportedTypes = { Boolean().encode_tag(): Boolean,
                   Integer().encode_tag(): Integer,
                   BitString().encode_tag(): BitString,
                   OctetString().encode_tag(): OctetString,
                   Null().encode_tag(): Null,
                   ObjectIdentifier().encode_tag(): ObjectIdentifier,
                   Real().encode_tag(): Real,
                   Enumerated().encode_tag(): Enumerated,
                   Sequence().encode_tag(): Sequence,
                   SequenceOf().encode_tag(): SequenceOf,
                   Set().encode_tag(): Set,
                   SetOf().encode_tag(): SetOf }

def decode(input):
    """
       decode(input) -> (asn1, rest)
       
       Decode input octet stream (string) into ASN.1 object and return
       the rest of input (string).
    """
    if type(input) != StringType or len(input) == 0:
        raise error.BadEncodingError('Malformed BER octet stream')

    if not supportedTypes.has_key(input[0]):
        raise error.BadEncodingError('Unknown BER tag %s' % repr(input[0]))

    if not supportedTypes.has_key(input[0]):
        # Unknown tag
        return (None, input)

    obj = supportedTypes[input[0]]()
    return (obj, obj.decode(input)[1])
