"""
   ASN.1 "universal" data types.

   Copyright 1999-2002 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
# Module public names
__all__ = [ 'Boolean', 'Integer', 'BitString', 'OctetString', 'Null', \
            'ObjectIdentifier', 'Real', 'Enumerated', 'Sequence', \
            'SequenceOf', 'Set', 'SetOf', 'Choice' ]

from operator import getslice, setslice, delslice
from types import IntType, LongType, StringType, NoneType, FloatType,  \
     ListType, SliceType
from string import split, atol, atol_error
from pysnmp.asn1 import base, error

#
# ASN.1 "simple" types implementation
#

class Boolean(base.SimpleAsn1Object):
    """An ASN.1 boolean object
    """
    tagId = 0x01
    allowedTypes = [ IntType, LongType ]
    singleValueConstraint = [ 0, 1 ]
    initialValue = 0L

    # Disable not applicible constraints
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

    # Basic logical ops
    
    def __and__(self, value):
        """Perform binary AND operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() & value.get())

    __rand__ = __and__

    def __or__(self, value):
        """Perform binary OR operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() | value.get())

    __ror__ = __or__

    def __xor__(self, value):
        """Perform binary XOR operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() ^ value.get())

    __rxor__ = __xor__

    def __iand__(self, value):
        """Perform binary AND operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() & value.get())
        return self

    def __ior__(self, value):
        """Perform binary OR operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() | value.get())
        return self

    def __ixor__(self, value):
        """Perform binary XOR operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() ^ value.get())
        return self

class Integer(base.SimpleAsn1Object):
    """An ASN.1, indefinite length integer object
    """
    tagId = 0x02
    allowedTypes = [ IntType, LongType ]
    initialValue = 0L

    # Disable not applicible constraints
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

    # Basic arithmetic ops
    
    def __add__(self, value):
        """Add a value
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() + value.get())

    __radd__ = __add__
    
    def __sub__(self, value):
        """Subscract a value
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() - value.get())

    def __rsub__(self, value):
        """Subscract our value from given one
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(value.get() - self.get())
    
    def __mul__(self, value):
        """Multiply a value
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() * value.get())

    __rmul__ = __mul__
    
    def __div__(self, value):
        """Divide a value by ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() / value.get())

    def __rdiv__(self, value):
        """Divide ourselves by value
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(value.get() / self.get())

    def __mod__(self, value):
        """Take a modulo of ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() % value.get())

    __rmod__ = __mod__

    def __pow__(self, value, modulo):
        """Provision for pow()
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(pow(self.get(), value.get(), modulo))

    def __rpow__(self, value, modulo):
        """Provision for rpow()
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(pow(value.get(), self.get(), modulo))

    def __lshift__(self, value):
        """Perform left shift operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() << value.get())

    def __rshift__(self, value):
        """Perform right shift operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() >> value.get())

    def __and__(self, value):
        """Perform binary AND operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() & value.get())

    __rand__ = __and__

    def __or__(self, value):
        """Perform binary OR operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() | value.get())

    __ror__ = __or__

    def __xor__(self, value):
        """Perform binary XOR operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        return self.__class__(self.get() ^ value.get())

    __rxor__ = __xor__

    def __iadd__(self, value):
        """Add value to ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() + value.get())
        return self
    
    def __isub__(self, value):
        """Subscract value from ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() - value.get())
        return self

    def __imul__(self, value):
        """Multiply a value to ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() * value.get())
        return self
    
    def __idiv__(self, value):
        """Divide a value by ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() / value.get())
        return self

    def __imod__(self, value):
        """Take a modulo of ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() % value.get())
        return self

    def __ipow__(self, value, modulo):
        """Provision for x**=y
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(pow(self.get(), value.get(), modulo))
        return self

    def __ilshift__(self, value):
        """Perform left shift operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() << value.get())
        return self
    
    def __irshift__(self, value):
        """Perform right shift operation
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() >> value.get())
        return self

    def __iand__(self, value):
        """Perform binary AND operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() & value.get())
        return self

    def __ior__(self, value):
        """Perform binary OR operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() | value.get())
        return self

    def __ixor__(self, value):
        """Perform binary XOR operation against ourselves
        """
        if not isinstance(value, self.__class__):
            value = self.__class__(value)
        self.set(self.get() ^ value.get())
        return self

    def __int__(self):
        """Return an integer value of ourselves
        """
        return int(self.get())

    def __long__(self):
        """Return a long integer value of ourselves
        """
        return long(self.get())

    def __float__(self):
        """Return a floating point value of ourselves
        """
        return float(self.get())    

class BitString(base.SimpleAsn1Object):
    """An ASN.1 BITSTRING object XXX
    """
    tagId = 0x03
    allowedTypes = [ StringType ]

    # Disable not applicible constraints
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

class OctetString(base.SimpleAsn1Object):
    """ASN.1 octet string object
    """
    tagId = 0x04
    allowedTypes = [ StringType ]
    initialValue = ''
    
    # Disable not applicible constraints
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

    # Immutable sequence object protocol
    
    def __len__(self): return len(self.get())
    def __getitem__(self, i):
        """Get string component by index or slice
        """
        if type(i) == SliceType:
            return self.__class__(getslice(self.get(), i.start, i.stop))
        else:
            return self.get()[i]

    def __add__(self, other):
        """Add sub-id  with input verification
        """
        val = self.get() + self.__class__(other).get()
        return self.__class__(val)

    def __radd__(self, other):
        """Add sub-id  with input verification
        """
        val = list(self.__class__(other)) + self._getraw()
        return self.__class__(val) 

    def __mul__(self, value):
        """Multiply a value
        """
        return self.__class__(self.get() * value)

    __rmul__ = __mul__

    # They won't be defined if version is at least 2.0 final
    def __getslice__(self, i, j):
        return self[max(0, i):max(0, j):]

class Null(base.SimpleAsn1Object):
    """ASN.1 NULL object
    """
    tagId = 0x05
    allowedTypes = [ IntType, LongType, StringType, NoneType ]
    singleValueConstraint = [ 0, 0L, '', None ]

    # Disable not applicible constraints
    _subtype_contained_subtype_constraint = base.Asn1Object._subtype_contained_subtype_constraint_na
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

class ObjectIdentifier(base.SimpleAsn1Object):
    """ASN.1 Object ID object (taken and returned as string in conventional
       "dotted" representation)
    """
    tagId = 0x06
    allowedTypes = [ StringType, ListType ]
    initialValue = []
    
    # Disable not applicible constraints
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na
    
    # Sequence object protocol
    
    def __len__(self): return len(self._getraw())
    def __getitem__(self, i):
        """Get sequence component by index or slice
        """
        if type(i) == SliceType:
            return self.__class__(getslice(self._getraw(), i.start, i.stop))
        else:
            return self._getraw()[i]

    def __delitem__(self, i):
        """Delete sequence component by index or slice
        """
        if type(i) == SliceType:
            delslice(self._getraw(), i.start, i.stop)
        else:
            del self._getraw()[i]

    def __setitem__(self, i, item):
        """Set sub-id by index or slice with input verification
        """
        val = list(self);
        if type(i) == SliceType:
            setslice(val, i.start, i.stop, list(self.__class__(item)))
        else:
            val[i] = item
        self.set(self.__class__(val))

    def __add__(self, other):
        """Add sub-id  with input verification
        """
        val = self._getraw() + list(self.__class__(other))
        return self.__class__(val)

    def __radd__(self, other):
        """Add sub-id  with input verification
        """
        val = list(self.__class__(other)) + self._getraw()
        return self.__class__(val) 

    # They won't be defined if version is at least 2.0 final
    def __getslice__(self, i, j):
        return self[max(0, i):max(0, j):]
    def __setslice__(self, i, j, seq):
        self[max(0, i):max(0, j):] = seq
    def __delslice__(self, i, j):
        del self[max(0, i):max(0, j):]

    def append(self, item):
        """Append sub-id with input verification
        """
        val = list(self); val.append(item)
        self.set(self.__class__(val))

    def extend(self, item):
        """Append an oid with input verification
        """
        val = list(self); val.extend(list(self.__class__(item)))
        self.set(self.__class__(val))

    def index(self, suboid):
        """Returns index of first occurrence of given sub-identifier
        """
        val = list(self)
        return val.index(suboid)

    def isaprefix(self, other):
        """
           isaprefix(other) -> boolean
           
           Compare our own OID with the other one (given as a string),
           return non-None if ours is a prefix of the other OID.

           This is intended to be used for MIB tables retrieval.
        """
        # Normalize foreign value type to ours
        if not isinstance(other, self.__class__):
            other = self.__class__(other)
                        
        # Pick the shortest oid
        if len(self) <= len(other.value):
            # Get the length
            length = len(self)

            # Compare oid'es
            if self[:length] == other[:length]:
                return not None

        # Our OID turned to be greater than the other
        return None

    def str2num(self, soid):
        """
            str2num(soid) -> noid
            
            Convert Object ID "soid" presented in a dotted form into an
            Object ID "noid" represented as a list of numeric sub-ID's.
        """
        if not soid:
            return []
        
        # Convert string into a list and filter out empty members
        # (leading dot causes this)
        try:
            toid = filter(lambda x: len(x), split(soid, '.'))

        except:
            raise error.BadArgumentError('Malformed Object ID %s for %s' %\
                                         (str(soid), self.__class__.__name__))

        # Convert a list of symbols into a list of numbers
        try:
            noid = map(lambda x: atol(x), toid)

        except atol_error:
            raise error.BadArgumentError('Malformed Object ID %s for %s' %\
                                         (str(soid), self.__class__.__name__))

        if not noid:
            raise error.BadArgumentError('Empty Object ID %s for %s' %\
                                         (str(soid), self.__class__.__name__))

        return noid

    def num2str(self, noid):
        """
            num2str(noid) -> snoid
            
            Convert Object ID "noid" presented as a list of numeric
            sub-ID's into Object ID "soid" in dotted notation.
        """    
        if not noid:
            return ''

        # Convert a list of number into a list of symbols and merge all
        # list members into a string
        try:
            soid = reduce(lambda x, y: x+y,\
                          map(lambda x: '.%lu' % x, noid))
        except:
            raise error.BadArgumentError('Malformed numeric Object ID %s for %s' % (str(noid), self.__class__.__name__))
 
        if not soid:
            raise error.BadArgumentError('Empty numeric Object ID %s for %s' %\
                                         (str(noid), self.__class__.__name__))

        return soid

    def _iconv(self, value):
        """Input filter: accept both string & list type OIDs
        """
        if type(value) == StringType:
            return self.str2num(value)
        if type(value) == ListType:
            return self.str2num(self.num2str(value))

    _oconv = num2str
            
class Real(base.SimpleAsn1Object):
    """An ASN.1 REAL object XXX
    """
    tagId = 0x09
    allowedTypes = [ IntType, LongType, FloatType ]
    initialValue = 0.0

    # Disable not applicible constraints
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

class Enumerated(base.SimpleAsn1Object):
    """An ASN.1 ENUMERATED object XXX
    """
    tagId = 0x10
    allowedTypes = [ IntType, LongType ]
    initialValue = 0

    # Disable not applicible constraints
    _subtype_value_range_constraint = base.Asn1Object._subtype_value_range_constraint_na    
    _subtype_size_constraint = base.Asn1Object._subtype_size_constraint_na
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

#
# ASN.1 "structured" types implementation
#

class Sequence(base.RecordTypeAsn1Object):
    """ASN.1 sequence object
    """
    tagId = 0x10

class SequenceOf(base.VariableTypeAsn1Object):
    """ASN.1 Sequence Of object
    """
    tagId = 0x10

    # Disable not applicible constraints
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

class Set(base.RecordTypeAsn1Object):
    """ASN.1 Set object
    """
    tagId = 0x11

class SetOf(base.VariableTypeAsn1Object):
    """ASN.1 Set Of object
    """
    tagId = 0x11

    # Disable not applicible constraints
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na

class Choice(base.ChoiceTypeAsn1Object):
    """ASN.1 Choice clause
    """
    # Untagged type XXX
    tagCategory = base.tagCategories['UNTAGGED']
    
    # Disable not applicible constraints XXX
    _subtype_permitted_alphabet_constraint = base.Asn1Object._subtype_permitted_alphabet_constraint_na
