#!/usr/local/bin/python -O
"""
   Retrieve a SNMP table of Object IDs starting from user specifed SNMP
   Object ID(s) from arbitrary SNMP agent using SNMP v.2 GETBULK PDU.

   Since MIB parser is not yet implemented in Python, this script takes and
   reports Object IDs in dotted numeric representation only.

   Copyright 1999-2002 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
import sys, getopt
from pysnmp.proto import v1, v2c
from pysnmp.mapping.udp import role

# Initialize help messages
options =           'Options:\n'
options = options + '  -p <port>      port to communicate with at the agent. Default is 161.\n'
options = options + '  -r <retries>   number of retries to be used in requests. Default is 5.\n'
options = options + '  -t <timeout>   timeout between retries. Default is 1.\n'
options = options + '  -R             report variables types on output.'
usage = 'Usage: %s [options] <snmp-agent> <community> <obj-id [[obj-id] ... ]>\n' % sys.argv[0]
usage = usage + options
    
# Initialize defaults
port = 161; retries = 5; timeout = 1; reportTypeFlag = None

# Parse possible options
try:
    (opts, args) = getopt.getopt(sys.argv[1:], 'hp:r:t:R',\
                                 ['help', 'port=', 'retries=',\
                                  'timeout=', 'report-type'])
except getopt.error, why:
    print 'getopt error: %s\n%s' % (why, usage)
    sys.exit(-1)

try:
    for opt in opts:
        if opt[0] == '-h' or opt[0] == '--help':
            print usage
            sys.exit(0)
        
        if opt[0] == '-p' or opt[0] == '--port':
            port = int(opt[1])

        if opt[0] == '-r' or opt[0] == '--retries':
            retries = int(opt[1])

        if opt[0] == '-t' or opt[0] == '--timeout':
            timeout = int(opt[1])

        if opt[0] == '-R' or opt[0] == '--report-type':
            reportTypeFlag = 1

except ValueError, why:
    print 'Bad parameter \'%s\' for option %s: %s\n%s' \
          % (opt[1], opt[0], why, usage)
    sys.exit(-1)

if len(args) < 3:
    print 'Insufficient number of arguments supplied\n%s' % usage
    sys.exit(-1)

# Create SNMP manager object
client = role.manager((args[0], port))

# Pass it a few options
client.timeout = timeout
client.retries = retries

# Store tables headers
head_oids = args[2:]

# Initialize initial OID(s)
oids = head_oids

# Create SNMP GETBULK request
req = v2c.GetBulkRequest()

# Set community name to request
req['community'].set(args[1])

# Build and set Object ID & variables bindings
req['pdu']['get_bulk_request']['variable_bindings'].extend(map(lambda x: v2c.VarBind(name=v2c.ObjectName(x)), args[2:]))

# Create a response message framework
rsp = v2c.Response();

# Traverse agent MIB
while 1:
    # Prepare OID(s) for next GET request
    req['pdu'].values()[0]['variable_bindings']=v2c.VarBindList(\
            map(lambda x: v2c.VarBind(name=v2c.ObjectName(x)), oids))
    
    # Encode SNMP request message and try to send it to SNMP agent and
    # receive a response
    (answer, src) = client.send_and_receive(req.encode())
    
    # Attempt to decode SNMP response
    rsp.decode(answer)

    # Make sure response matches request
    if not req.match(rsp):
        raise 'Unmatched response: %s vs %s' % (req, rsp)

    # Fetch Object ID's and associated values
    oids = map(lambda x: x['name'].get(), \
               rsp['pdu'].values()[0]['variable_bindings'])
    vals = map(lambda x: x['value'], \
               rsp['pdu'].values()[0]['variable_bindings'])

    # Check for remote SNMP agent failure
    if rsp['pdu'].values()[0]['error_status']:
        raise str(rsp['pdu'].values()[0]['error_status']) + ' at '\
              + str(oids[rsp['pdu'].values()[0]['error_index'].get()-1])

    # The following is taken from RFC1905 (fixed not to depend of repetitions)
    N = 0;
    R = len(req['pdu'].values()[0]['variable_bindings']) - N
    M = len(rsp['pdu'].values()[0]['variable_bindings']) / R
    for i in range(1, M+1):
        for r in range(1, R+1):
            idx = N + ((i-1)*R) + r
            print oids[idx-1], ' ---> ',
            if reportTypeFlag:
                print vals[idx-1]
            else:
                print repr(vals[idx-1].get())

    # Leave the last of each requested OIDs
    oids = oids[-R:]
    vals = vals[-R:]

    # Exclude completed OIDs
    for idx in range(len(vals)):
        if vals[idx].has_key('endOfMibView'):
            del oids[idx]
    
    if not oids:
        sys.exit(0)
        
    # Update request ID
    req['pdu'].values()[0]['request_id'] = req['pdu'].values()[0]['request_id'] + 1
