#!/usr/local/bin/python -O
"""
   Set new values to MIB variables associated with user specifed
   SNMP Object IDs at arbitrary SNMP agent.

   Since MIB parser is not yet implemented in Python, this script takes and
   reports Object IDs in dotted numeric representation only.

   Copyright 1999-2002 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
import sys, getopt
from pysnmp.proto import v1, v2c
from pysnmp.mapping.udp import role

# Initialize help messages
options =           'Options:\n'
options = options + '  -p <port>      port to communicate with at the agent. Default is 161.\n'
options = options + '  -r <retries>   number of retries to be used in requests. Default is 5.\n'
options = options + '  -t <timeout>   timeout between retries. Default is 1.\n'
options = options + '  -v <version>   SNMP version to use [1, 2c]. Default is 1 (version one).\n'
options = options + '  -R             report variables types on output.\n'
types =         'Types:\n'
types = types + '   i:            INTEGER\n'
types = types + '   u:            unsigned INTEGER\n'
types = types + '   t:            TIMETICKS\n'
types = types + '   a:            IPADDRESS\n'
types = types + '   o:            OBJID\n'
types = types + '   s:            STRING\n'
types = types + '   U:            COUNTER64 (version 2 and above)'
usage = 'Usage: %s [options] <snmp-agent> <community> <oid type val [...]>\n' % sys.argv[0]
usage = usage + options + types
    
# Initialize defaults
port = 161; retries = 5; timeout = 1; version = '1'; reportTypeFlag = None
    
# Parse possible options
try:
    (opts, args) = getopt.getopt(sys.argv[1:], 'hp:r:t:v:R',\
                                 ['help', 'port=', 'retries=', \
                                  'timeout=', 'version=', 'report-type'])

except getopt.error, why:
    print 'getopt error: %s\n%s' % (why, usage)
    sys.exit(-1)

try:
    for opt in opts:
        if opt[0] == '-h' or opt[0] == '--help':
            print usage
            sys.exit(0)

        if opt[0] == '-p' or opt[0] == '--port':
            port = int(opt[1])

        if opt[0] == '-r' or opt[0] == '--retries':
            retries = int(opt[1])

        if opt[0] == '-t' or opt[0] == '--timeout':
            timeout = int(opt[1])

        if opt[0] == '-v' or opt[0] == '--version':
            version = opt[1]

        if opt[0] == '-R' or opt[0] == '--report-type':
            reportTypeFlag = 1

except ValueError, why:
    print 'Bad parameter \'%s\' for option %s: %s\n%s' \
          % (opt[1], opt[0], why, usage)
    sys.exit(-1)

if len(args) < 5:
    print 'Insufficient number of arguments supplied\n%s' % usage
    sys.exit(-1)

# Choose protocol version specific module
try:
    snmp = eval('v' + version)

except (NameError, AttributeError):
    print 'Unsupported SNMP protocol version: %s\n%s' % (version, usage)
    sys.exit(-1)

# Try to provide usage-level compatibility with UCD snmpset
idx = 2; bindings=snmp.VarBindList()
while idx < len(args)-2:
    try:
        if args[idx+1] == 'i':
            val = snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.Integer(long(args[idx+2]))))
        elif args[idx+1] == 'u':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.Unsigned32(long(args[idx+2]))))
        elif args[idx+1] == 't':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.TimeTicks(long(args[idx+2]))))
        elif args[idx+1] == 'a':
            if version == '1':
                val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.NetworkAddress(internet=snmp.IpAddress(args[idx+2]))))
            else:
                val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.IpAddress(args[idx+2])))
        elif args[idx+1] == 'o':
            val=snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.ObjectIdentifier(args[idx+2])))
        elif args[idx+1] == 's':
            val = snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.OctetString(args[idx+2])))
        elif args[idx+1] == 'U' and version != '1':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.Counter64(long(args[idx+2]))))
        else:
            print 'Unknown value type \'%s\'\n%s' % (args[idx+1], usage)
            sys.exit(-1)

        if version == '1':
            bindings.append(snmp.VarBind(\
                name=snmp.ObjectName(args[idx]),\
                value=val))            
        else:            
            bindings.append(snmp.VarBind(\
                name=snmp.ObjectName(args[idx]),\
                value=snmp.BindValue(value=val)))

        idx = idx + 3
            
    except IndexError:
        print 'Wrong number of arguments supplied\n%s' % usage
        sys.exit(-1)

    except (TypeError, ValueError), why:
        print 'Wrong type of value \'%s\': %s' % (args[idx+2], why)
        sys.exit(-1)

    except AttributeError, why:
        print 'Unsupported ASN.1 type %s for protocol version %s' %\
              (why, version)
        sys.exit(-1)
        
# Create SNMP manager object
client = role.manager((args[0], port))

# Pass it a few options
client.timeout = timeout
client.retries = retries

# Create SNMP SET request
req = snmp.SetRequest()

# Set community name to request
req['community'].set(args[1])

# Build and set Object ID & variables bindings
req['pdu']['set_request']['variable_bindings'] = bindings

# Encode SNMP request message and try to send it to SNMP agent and
# receive a response
(answer, src) = client.send_and_receive(req.encode())

# Decode SNMP response
rsp = snmp.GetResponse(); rsp.decode(answer)

# Make sure response matches request
if not req.match(rsp):
    raise 'Unmatched response: %s vs %s' % (req, rsp)

# Fetch Object ID's and associated values
oids = map(lambda x: x['name'].get(), \
           rsp['pdu'].values()[0]['variable_bindings'])
vals = map(lambda x: x['value'], \
           rsp['pdu'].values()[0]['variable_bindings'])

# Check for remote SNMP agent failure
if rsp['pdu'].values()[0]['error_status']:
    raise str(rsp['pdu'].values()[0]['error_status']) + ' at '\
          + str(oids[rsp['pdu'].values()[0]['error_index'].get()-1])

for (oid, val) in map(None, oids, vals):
    print oid, ' ---> ',
    if reportTypeFlag:
        print val
    else:
        print repr(val.get())
