#!/usr/local/bin/python -O
"""
   Send SNMP trap to remote SNMP manager.

   Since MIB parser is not yet implemented in Python, this script takes and
   reports Object IDs in dotted numeric representation only.

   Copyright 1999-2002 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
import sys, getopt
from pysnmp.proto import v1, v2c
from pysnmp.mapping.udp import role

# Initialize help messages
options =           'Options:\n'
options = options + '  -p <port>      port to communicate with at the agent. Default is 162.\n'
options = options + '  -v <version>   SNMP version to use [1, 2c]. Default is 1 (version one).\n'
types =         'Types:\n'
types = types + '   i:            INTEGER\n'
types = types + '   u:            unsigned INTEGER\n'
types = types + '   t:            TIMETICKS\n'
types = types + '   a:            IPADDRESS\n'
types = types + '   o:            OBJID\n'
types = types + '   s:            STRING\n'
types = types + '   U:            COUNTER64 (version 2 and above)'
usage = 'Usage: %s [options] <snmp-agent> <community> [<trap-parameters>]\n' % sys.argv[0]
trapopt =           '-v 1 trap parameters:\n'
trapopt = trapopt + '   [enterprise-oid [agent [trap-type [specific-type [uptime [oid <type val> [...]]]]]]]\n'
trapopt = trapopt + '-v 2c trap parameters:\n'
trapopt = trapopt + '   [uptime [trap-oid [<oid type val> [...]]]]\n'
usage = usage + options + trapopt + types
    
# Initialize defaults
port = 161
version = '1'
    
# Parse possible options
try:
    (opts, args) = getopt.getopt(sys.argv[1:], 'hp:v:',\
                                 ['help', 'port=', 'version='])
except getopt.error, why:
    print 'getopt error: %s\n%s' % (why, usage)
    sys.exit(-1)

try:
    for opt in opts:
        if opt[0] == '-h' or opt[0] == '--help':
            print usage
            sys.exit(0)
        
        if opt[0] == '-p' or opt[0] == '--port':
            port = int(opt[1])

        if opt[0] == '-v' or opt[0] == '--version':
            version = opt[1]

except ValueError, why:
    print 'Bad parameter \'%s\' for option %s: %s\n%s' \
          % (opt[1], opt[0], why, usage)
    sys.exit(-1)

# Choose protocol version specific module
try:
    snmp = eval('v' + version)

except (NameError, AttributeError):
    print 'Unsupported SNMP protocol version: %s\n%s' % (version, usage)
    sys.exit(-1)

if len(args) < 2:
    print 'Insufficient number of arguments supplied\n%s' % usage
    sys.exit(-1)

agent = args[0]

# Create SNMP Trap request
if version == '1':
    req = snmp.Trap()
else:
    req = snmp.SnmpV2Trap()

req['community'].set(args[1])

bindings=snmp.VarBindList()

# Load trap params
if version == '1':
    if len(args) > 2 and args[2]:
        req['pdu']['trap']['enterprise'].set(args[2])
    if len(args) > 3 and args[3]:
        req['pdu']['trap']['agent_addr']['internet'].set(args[3])        
    else:
        import socket
        req['pdu']['trap']['agent_addr']['internet'].set(\
            socket.gethostbyname(socket.gethostname()))
    if len(args) > 4 and args[4]:
        req['pdu']['trap']['generic_trap'].set(int(args[4]))        
    if len(args) > 5 and args[5]:
        req['pdu']['trap']['specific_trap'].set(int(args[5]))
    if len(args) > 6 and args[6]:
        req['pdu']['trap']['time_stamp'].set(int(args[6]))        
    args = args[7:]
elif version == '2c':
    # Our uptime
    if len(args) > 2 and args[2]:
        uptime = int(args[2])
    else:
        from time import time
        uptime = int(time())

    # TrapOID
    if len(args) > 3 and args[3]:
        trapoid = args[3]
    else:
        trapoid = '1.3.6.1.4.1.3.1.1'

    args = args[4:]
    
    # First add mandatory variable bindings (see RFC1905)
    bindings.extend(map(lambda x, y: snmp.VarBind(\
        name=snmp.ObjectName(x),
        value=snmp.BindValue(value=snmp.ObjectSyntax(syntax=y))),
        [ '.1.3.6.1.2.1.1.3.0', '.1.3.6.1.6.3.1.1.4.1.0' ],
        [ snmp.ApplicationSyntax(value=snmp.TimeTicks(uptime)),
          snmp.SimpleSyntax(value=snmp.ObjectIdentifier(trapoid))]))

# Parse and add the rest of trap OIDs & vals (see RFC1905)
idx = 0
while idx < len(args)-2:
    try:
        if args[idx+1] == 'i':
            val = snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.Integer(long(args[idx+2]))))
        elif args[idx+1] == 'u':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.Unsigned32(long(args[idx+2]))))
        elif args[idx+1] == 't':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.TimeTicks(long(args[idx+2]))))
        elif args[idx+1] == 'a':
            if version == '1':
                val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.NetworkAddress(internet=snmp.IpAddress(args[idx+2]))))
            else:
                val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.IpAddress(args[idx+2])))
        elif args[idx+1] == 'o':
            val=snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.ObjectIdentifier(args[idx+2])))
        elif args[idx+1] == 's':
            val = snmp.ObjectSyntax(syntax=snmp.SimpleSyntax(value=snmp.OctetString(args[idx+2])))
        elif args[idx+1] == 'U' and version != '1':
            val = snmp.ObjectSyntax(syntax=snmp.ApplicationSyntax(value=snmp.Counter64(long(args[idx+2]))))
        else:
            print 'Unknown value type \'%s\'\n%s' % (args[idx+1], usage)
            sys.exit(-1)

        if version == '1':
            bindings.append(snmp.VarBind(\
                name=snmp.ObjectName(args[idx]),\
                value=val))            
        else:            
            bindings.append(snmp.VarBind(\
                name=snmp.ObjectName(args[idx]),\
                value=snmp.BindValue(value=val)))

        idx = idx + 3

    except IndexError:
        print 'Wrong number of arguments supplied\n%s' % usage
        sys.exit(-1)

    except (TypeError, ValueError), why:
        print 'Wrong type of value \'%s\': %s' % (args[idx+2], why)
        sys.exit(-1)

req['pdu'].values()[0]['variable_bindings'] = bindings

# Create SNMP manager object
client = role.manager((agent, port))

# Encode SNMP request message and try to send it to SNMP manager
client.send(req.encode())

